@c -*-texinfo-*-

@node Minimum Spanning Tree Computation, Priority Queues, Uniformity Testing, Top
@chapter Minimum Spanning Tree Computation

Computing the MST of the set of pooled points is the most interesting
part of the implementation of the uniformity test.  It is also the part
that offers the most possibilities.  Several methods for computing a
minimum spanning tree are suitable.  The following table lists the ones
considered, along with their asymptotic time requirements for a complete
graph with |n| vertices:

@multitable @columnfractions .05 .45 .21 .2
@item @tab Method
@tab Best/Avg. Time
@tab Worst Time

@item @tab Kruskal's algorithm
@tab O(@altmath{n^2 \log n, n**2 lg n})
@tab O(@altmath{n^2 \log n, n**2 lg n})

@item @tab Prim's algorithm with binary heap
@tab O(@altmath{n^2 \log n, n**2 lg n})
@tab O(@altmath{n^2 \log n, n**2 lg n})

@item @tab Prim's algorithm with Fibonacci heap
@tab O(@altmath{n^2, n**2})
@tab O(@altmath{n^2, n**2})

@item @tab Bentley's single-fragment algorithm
@tab O(@altmath{n \log n, n lg n})
@tab O(@altmath{n^2 \log n, n**2 lg n})

@item @tab Bentley's multifragment algorithm
@tab O(@altmath{n \log n, n lg n})
@tab (unknown)
@end multitable

Each of these algorithms has its own advantages and disadvantages in
this context.  Kruskal's algorithm is simple to implement, but slow.
Prim's algorithm requires a priority queue.  If a binary heap is used
for the priority queue, then it has the same asymptotic speed as
Kruskal's algorithm.  Substituting a Fibonacci heap speeds up processing
considerably.  Bentley's single-fragment algorithm is faster than any
other algorithm in the average case, but in the worst case it degrades
to worse than Prim's using a Fibonacci heap.  Unfortunately, this worst
case occurs when the points are clustered, one of the situations of
interest to us.  Bentley's multifragment algorithm may remedy this
problem, but its worst-case time bound is in fact unknown.

Prim's algorithm was chosen for implementation because it has the best
worst-case behavior of any of the algorithms.  Both binary heap and
Fibonacci heap variants are implemented.  The object-oriented
architecture used allows different algorithms, such as Bentley's, to be
substituted later or even chosen dynamically at runtime.

@references
@bibref{Fredman 1987};
@bibref{Cormen 1990}, chapters 21 and 24;
@bibref{Bentley 1978}.

@menu
* MST Class::                   
* The Algorithm of Prim::       
* MST Class Implementations::   
@end menu

@node MST Class, The Algorithm of Prim, Minimum Spanning Tree Computation, Minimum Spanning Tree Computation
@section Class Structure

The ``class'' structure that allows alternative MST implementations to
be used is |struct unf_mst|.  Its only member is pointer to function
|unf_run|.  This function takes as arguments a memory allocator class, a
pointer to a set of points, and counts of the points and the number of
dimensions that they contain.  It returns a pointer to pairs of vertex
indexes indicating the edges in the MST.

@<Test options@> +=
typedef int unf_mst_result[2];

struct unf_mst @
  {@-
    unf_mst_result *(*unf_run) (struct unf_mem *, @
                                const double *, int, int);
  };@+

@

The Prim's algorithm MST implemented here uses an additional class
structure |struct pq_class|, which represents a particular type of
priority queue.  This structure will be presented later, along with the
priority queue implementations themselves.

@node The Algorithm of Prim, MST Class Implementations, MST Class, Minimum Spanning Tree Computation
@section Prim's Algorithm

Prim's algorithm for computing a minimum spanning tree is implemented by
@file{mst-prim.c}.  It begins as follows:

@(mst-prim.c@> =
@<BSD License@>

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include "pq.h"
#include "uniformity.h"

@

The first piece of code is a helper function to calculate the Euclidean
distance between two points:

@(mst-prim.c@> +=
/* Calculates and returns the distance between points |a| and |b|, @
   each in |d| dimensions. */
static double @
calc_distance (const double *a, const double *b, int d) @
{
  double sum = 0.0;

  assert (a != NULL && b != NULL && d >= 0);
  while (d--) @
    {@-
      double diff = *a++ - *b++;
      sum += diff * diff;
    }@+

  return sum;
}

@

The MST function itself is next.  Its outline looks like this:

@(mst-prim.c@> +=
/* Calculates the MST of the |n| |d|-dimensional points at |p|.
   Uses |mem| for memory allocation
   and an instance of |pq_class| as a priority queue. */
static unf_mst_result *@
calc_mst (struct pq_class *pq_class,
	  struct unf_mem *mem, const double *p, int n, int d) @
{
  struct pq *pq;
  unf_mst_result *mst;
  int *edges;

  assert (pq_class != NULL && mem != NULL && p != NULL @
	  && n > 0 && d > 0);

  @<Memory allocation for Prim MST@>
  @<Execute Prim MST@>
  @<Record results of Prim MST@>
  @<Clean up and return from Prim MST@>
}

@

The first step is memory allocation.  We need to create a priority
queue, temporary space used for storing edges, and space in which to
return the result to the user:

@<Memory allocation for Prim MST@> =
pq = pq_class->create (mem, n);
mst = mem->unf_alloc (sizeof *mst * (n - 1));
edges = mem->unf_alloc (sizeof *edges * n);
if (pq == NULL || mst == NULL || edges == NULL) @
  {@-
    if (pq != NULL)
      pq_class->discard (pq);
    mem->unf_free (mst);
    mem->unf_free (edges);
    return NULL;
  }@+

@

The actual computation of the MST is the second step.  When this step is
complete, array |edges[]| implicitly stores the minimum spanning tree:
for each |i| in 1@dots{}n, there is an edge between vertices |i| and
|edges[i]|.  The first element, |edges[0]|, is not used.

@<Execute Prim MST@> =
pq_class->decrease_key (pq, 0, 0);
while (pq_class->count (pq) > 0) @
  {@-
    int u, v;

    u = pq_class->extract_min (pq);
    for (v = 0; v < n; v++)
      if (u != v) @
        {@-
          double key = pq_class->key (pq, v);

          if (key != -DBL_MAX) @
            {@-
              double cost = calc_distance (p + d * u, p + d * v, d);

              if (cost < key) @
                {@-
                  edges[v] = u;
                  pq_class->decrease_key (pq, v, cost);
                }@+
            }@+
        }@+
  }@+

@

After executing the MST, we translate the |edges[]| array into the form
expected by the caller.

@<Record results of Prim MST@> =
{
  int i;

  for (i = 1; i < n; i++) @
    {@-
      mst[i - 1][0] = i;
      mst[i - 1][1] = edges[i];
    }@+
}
@

Finally, we free temporary data and return to the caller.

@<Clean up and return from Prim MST@> =
pq_class->discard (pq);
mem->unf_free (edges);

return mst;
@

@references
@bibref{Cormen 1990}, section 24.2.

@node MST Class Implementations,  , The Algorithm of Prim, Minimum Spanning Tree Computation
@section Class Implementations

The final task is to declare class structures for the two MSTs defined
here.  This requires a wrapper function for each MST, because
|calc_mst()| requires one more argument than does |struct unf_mst|'s
|unf_run| member.  In all, we have the following:

@(mst-prim.c@> +=
/* Prim's Algorithm MST with binary heap priority queue. */
static unf_mst_result *@
calc_mst_prim_bin_heap (struct unf_mem *mem, const double *p, int n, int d)
{
  extern struct pq_class unf_pq_bin_heap;

  return calc_mst (&unf_pq_bin_heap, mem, p, n, d);
}

struct unf_mst unf_mst_prim_binary = @
  { @
    calc_mst_prim_bin_heap, @
  };

/* Prim's Algorithm MST with Fibonacci heap priority queue. */
static unf_mst_result *@
calc_mst_prim_fib_heap (struct unf_mem *mem, const double *p, int n, int d)
{
  extern struct pq_class unf_pq_fib_heap;

  return calc_mst (&unf_pq_fib_heap, mem, p, n, d);
}

struct unf_mst unf_mst_prim_fibonacci = @
  { @
    calc_mst_prim_fib_heap, @
  };
@
