@c -*-texinfo-*-

@node Other Classes, Discussion, Priority Queues, Top
@chapter Other Classes

This chapter contains descriptions and implementations of the three
classes that have not been treated yet: memory allocation, random number
generation, and point set handling.

@menu
* Memory Allocation::           
* Random Number Generator::     
* Point Set::                   
@end menu

@node Memory Allocation, Random Number Generator, Other Classes, Other Classes
@section Memory Allocation

|struct unf_mem| provides a ``class'' structure for memory allocators
within the uniformity testing library.  It is declared as follows:

@<Test options@> +=
/* Memory allocator class. */
struct unf_mem @
  {@-
    void *(*unf_alloc) (size_t);
    void (*unf_free) (void *);
  };@+

@

@noindent
In an implementation, |unf_alloc| must be a function that takes a count
of bytes to allocate and returns a pointer to the allocated memory if
successful or |NULL| if no memory is available.  Companion member
|unf_free| must point to a function that frees memory allocated with
|unf_alloc|.

File @(mem-std.c@> provides an implementation of the memory allocator
class that uses the standard C |malloc()| and |free()| functions:

@(mem-std.c@> =
@<BSD License@>

#include <stdlib.h>
#include "uniformity.h"

static void *@
standard_alloc (size_t n) @
{
  return malloc (n);
}

static void @
standard_free (void *p) @
{
  free (p);
}

struct unf_mem unf_mem_malloc = @
  {@
    standard_alloc, @
    standard_free@
  };
@

@node Random Number Generator, Point Set, Memory Allocation, Other Classes
@section Random Number Generator

The class interface for random number generation contains one important
member |unf_get|, which, when called, must return a pseudo-random number
in the closed range [0,@w{ }1].  An additional member |unf_seed| is provided
for seeding the generator for the convenience of calling code, but it is
not used by the uniformity testing library itself.

@<Test options@> +=
struct unf_rng @
  {@-
    void (*unf_seed) (unsigned long);
    double (*unf_get) (void);
  };@+

@

File @(rng-std.c@> uses the standard C pseudo-random number generator
(PRNG) |rand()| for random number generation.  This is not a good choice
on many systems, where |rand()| may have very poor randomness
properties.

@(rng-std.c@> =
@<BSD License@>

#include <stdlib.h>
#include "uniformity.h"

static void @
rng_seed (unsigned long seed) @
{
  srand (seed);
}

static double @
rng_get (void) @
{
  return (double) rand () / RAND_MAX;
}

struct unf_rng unf_rng_system = @
  {@
    rng_seed, @
    rng_get@
  };
@

An additional random number generator is provided in @file{rng-mt.c}.
It is based on the fast and high-quality Mersenne Twister random number
generator.

@references
@bibref{Matsumoto 1997}.

@node Point Set,  , Random Number Generator, Other Classes
@section Point Set

The purpose of the point set class is to represent an infinite set of
points that includes the convex hull of a provided finite collection of
points, but may include other points as well.  For instance, it could be
implemented as the smallest hypersphere or hyper-rectangle containing
all the provided points.  The class is represented as |struct unf_set|:

@<Test options@> +=
/* Point set. */
struct unf_set @
  {@-
    struct unf_a_set *(*unf_create) (struct unf_mem *, @
                                     const double *, int n, int d);
    void (*unf_discard) (struct unf_a_set *);
    void (*unf_random) (const struct unf_a_set *, @
                        struct unf_rng *, double *);
  };@+
@

The members of |struct unf_set| are abstractly described as follows:

@table @asis
@item |struct unf_a_set *unf_create (struct unf_mem *mem, const double *p, int n, int d)|
Creates and returns a new point set that includes the |n|
|d|-dimensional points at |p|.  Memory allocation is performed using
|mem|.  Returns |NULL| if unsuccessful due to lack of memory or for some
other reason.

@item |void unf_discard (struct unf_a_set *set);|
Destroys point set |set|.

@item |void unf_random (const struct unf_a_set *set, struct unf_rng *rng, double *p);|
Selects a random point from a uniform distribution over the set of
points represented by |set|.  Pseudo-random number generator |rng| is
used in selection of the point.  The point is stored into the vector at
|p|, which must have room for |d| elements, |d| being the same value
passed to |unf_create()|.
@end table

The implementation provided here in @(set-rect.c@> uses a minimum-size
hyper-rectangle to surround the points.  The implementation begins with
the definition of |struct unf_a_set|:

@(set-rect.c@> =
@<BSD License@>

#include <assert.h>
#include "uniformity.h"

/* Hyper-rectangular point set. */
struct unf_a_set @
  {@-
    struct unf_mem mem; /* Memory allocator. */

    int d;              /* Number of dimensions. */
    double *min;        /* Minimum values for each of |d| dimensions. */
    double *max;        /* Maximum values for each of |d| dimensions. */
  };@+
@

To create a set, we allocate memory for it, then iterate through the
set of points provided setting up the |min| and |max| arrays:

@(set-rect.c@> +=
struct unf_a_set *@
set_create (struct unf_mem *mem, const double *p, int n, int d) @
{
  struct unf_a_set *set;
  int i;

  assert (p != NULL && n > 0 && d > 0);

  /* Memory allocation. */
  set = mem->unf_alloc (sizeof *set);
  if (set == NULL)
    return NULL;
  set->mem = *mem;

  set->d = d;
  set->min = mem->unf_alloc (sizeof *set->min * d * 2);
  if (set->min == NULL) @
    {@-
      mem->unf_free (set->min);
      mem->unf_free (set);
      return NULL;
    }@+
  set->max = set->min + d;

  /* Construct hyper-rectangle. */
  for (i = 0; i < d; i++)
    set->min[i] = set->max[i] = *p++;

  for (i = 1; i < n; i++) @
    {@-
      int j;

      for (j = 0; j < d; j++) @
	{@-
	  if (*p < set->min[j])
	    set->min[j] = *p;
	  else if (*p > set->max[j])
	    set->max[j] = *p;

	  p++;
	}@+
    }@+

  return set;
}

@

Destroying a set is just a matter of freeing its memory:

@(set-rect.c@> +=
void @
set_discard (struct unf_a_set *set) @
{
  assert (set != NULL);

  set->mem.unf_free (set->min);
  set->mem.unf_free (set);
}
@

To generate a random point in a hyper-rectangle, we generate an uniform
random variate along each axis independently:

@(set-rect.c@> +=
void @
set_random (const struct unf_a_set *set, struct unf_rng *rng, double *v) @
{
  int i;
  
  for (i = 0; i < set->d; i++)
    v[i] = set->min[i] + rng->unf_get () * (set->max[i] - set->min[i]);
}
@

Finally, we declare the class structure implementation:

@(set-rect.c@> +=
struct unf_set unf_set_rectangular = @
  {@-
    set_create, @
    set_discard, @
    set_random,
  };@+
@
