/* Copyright (C) 2001 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@msu.edu> on the Internet,
   or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA through
   more mundane means.  
*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "pq.h"
#include "uniformity.h"
#include "test.h"

struct pq_bogo
  {
    int n;
    int m;

    double *key;
    int *vertex;
  };

const char *pgm_name = "pq-test";

static struct pq_bogo *
bogo_create (int n)
{
  struct pq_bogo *pq;
  int i;

  assert (n >= 0);
  pq = malloc (sizeof *pq);
  if (pq == NULL)
    return NULL;

  pq->key = malloc (sizeof *pq->key * n);
  pq->vertex = malloc (sizeof *pq->vertex * n);
  if (pq->key == NULL || pq->vertex == NULL)
    {
      free (pq->key);
      free (pq->vertex);
      free (pq);
      return NULL;
    }

  pq->n = pq->m = n;
  for (i = 0; i < n; i++)
    {
      pq->key[i] = DBL_MAX;
      pq->vertex[i] = i;
    }

  return pq;
}

static void
bogo_discard (struct pq_bogo *pq)
{
  assert (pq != NULL);

  free (pq->key);
  free (pq->vertex);
  free (pq);
}

static int
bogo_extract_min (struct pq_bogo *pq)
{
  int i;
  int min_index;
  int min_vertex;

  assert (pq != NULL);
  min_index = 0;
  for (i = 1; i < pq->n; i++)
    if (pq->key[i] < pq->key[min_index])
      min_index = i;
  min_vertex = pq->vertex[min_index];

  pq->n--;
  pq->key[min_index] = pq->key[pq->n];
  pq->vertex[min_index] = pq->vertex[pq->n];
  
  return min_vertex;
}

static void
bogo_decrease_key (struct pq_bogo *pq, int vertex, double key)
{
  int i;
  
  assert (pq != NULL);
  assert (vertex >= 0 && vertex < pq->m);
  
  for (i = 0; i < pq->n; i++)
    if (pq->vertex[i] == vertex)
      {
	pq->key[i] = key;
	return;
      }
}

/* Arrange the N elements of ARRAY in random order. */
void
shuffle (int *array, int n)
{
  int i;
  
  for (i = 0; i < n; i++)
    {
      int j = i + rand () % (n - i);
      int t = array[j];
      array[j] = array[i];
      array[i] = t;
    }
}

/* Command-line options. */
struct options
  {
    int count;
    int runs;

    enum queue
      {
	QUEUE_BIN_HEAP,
	QUEUE_FIB_HEAP
      }
    queue;

    unsigned seed;
  };
    
/* Print helpful syntax message and exit. */
static void 
usage (void) 
{
  printf ("pq-test, test program for priority queue implementations.\n\n"
	  "Usage: %s [OPTION]...\n\n"
	  "In the option descriptions below, CAPITALS denote arguments.\n"
	  "If a long option shows an argument as mandatory, then it is\n"
	  "mandatory for the equivalent short option also.\n\n"
	  "-c, --count=N         Number of items to test (default 128).\n"
	  "-q, --queue=MST       Priority queue to test:\n"
	  "                        bin-heap  Binary heap\n"
	  "                        fib-heap  Fibonacci heap\n"
	  "-r, --runs=RUNS       Number of runs (default 16).\n"
	  "-s, --seed=SEED       Initial random seed (time-based default).\n"
	  "-h, --help            Displays this help screen.\n"
	  "-V, --version         Reports version and copyright information.\n",
	  pgm_name);

  exit (EXIT_SUCCESS);
}

/* Parses command-line arguments from null-terminated array |args|.
   Sets up |options| appropriately to correspond. */
void
parse_command_line (char **args, struct options *options)
{
  static const struct long_option option_tab[] = 
    {
      {"count", 1, 'c'},
      {"queue", 1, 'q'},
      {"runs", 1, 'r'},
      {"seed", 1, 's'},
      {"help", 0, 'h'},
      {"version", 0, 'V'},
      {NULL, 0, 0},
    };

  struct option_state state;

  /* Default options. */
  options->count = 128;
  options->queue = QUEUE_BIN_HEAP;
  options->runs = 16;
  options->seed = time_seed () % 32768u;

  if (*args == NULL)
    return;

  option_init (&state, args + 1, "c:q:r:s:hV", option_tab);
  for (;;)
    {
      char *arg;
      int id = option_get (&state, &arg);
      if (id == -1)
	break;

      switch (id)
	{
	case 'c':
	  options->count = stoi (arg);
	  if (options->count < 1)
	    fail ("bad item count (%d)\n", options->count);
	  break;

	case 'q':
	  if (!strcmp (arg, "bin-heap"))
	    options->queue = QUEUE_BIN_HEAP;
	  else if (!strcmp (arg, "fib-heap"))
	    options->queue = QUEUE_FIB_HEAP;
	  else
	    fail ("unknown priority queue implementation `%s'", arg);
	  break;

	case 'r':
	  options->runs = stoi (arg);
	  if (options->runs < 1)
	    fail ("bad run count (%d)\n", options->runs);
	  break;

	case 's':
	  options->seed = stoi (arg);
	  break;

	case 'h':
	  usage ();

	case 'V':
	  fputs ("Priority queue unit test 1.0\n"
		 "Copyright (C) 2001 Ben Pfaff.\n"
		 "Copyright (C) 2001 Free Software Foundation, Inc.\n"
		 "See COPYING.BSD and COPYING.GPL for license.\n"
		 "This program comes with NO WARRANTY, not even for\n"
		 "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
		 stdout);
	  exit (EXIT_SUCCESS);

	default:
	  assert (0);
	}
    }
}

int
main (int argc, char *argv[])
{
  extern struct pq_class unf_pq_bin_heap;
  extern struct pq_class unf_pq_fib_heap;

  struct options cmdline;
  struct pq_class *class;
  int *vertex, *key;
  int i;
  
  parse_command_line (argv, &cmdline);
  switch (cmdline.queue)
    {
    case QUEUE_BIN_HEAP:
      class = &unf_pq_bin_heap;
      break;

    case QUEUE_FIB_HEAP:
      class = &unf_pq_fib_heap;
      break;

    default:
      assert (0);
    }

  vertex = malloc (sizeof *vertex * cmdline.count * 2);
  if (vertex == NULL)
    fail ("virtual memory exhausted");
  key = vertex + cmdline.count;

  for (i = 0; i < cmdline.runs; i++)
    {
      struct pq *a = class->create (&unf_mem_malloc, cmdline.runs + 1);
      struct pq_bogo *b = bogo_create (cmdline.runs);

      int j;

      printf ("Seed = %u...  ", cmdline.seed);
      fflush (stdout);
      srand (cmdline.seed);
      
      for (j = 0; j < cmdline.count; j++)
	vertex[j] = key[j] = j;
      shuffle (vertex, sizeof vertex / sizeof *vertex);
      shuffle (key, sizeof key / sizeof *key);

      class->decrease_key (a, cmdline.runs, 0);
      class->extract_min (a);

      for (j = 0; j < cmdline.runs; j++)
	{
	  class->decrease_key (a, vertex[j], key[j]);
	  bogo_decrease_key (b, vertex[j], key[j]);

	  class->verify (a);
	}

      for (j = 0; j < cmdline.runs; j++)
	{
	  int ma = class->extract_min (a);
	  int mb = bogo_extract_min (b);

	  class->verify (a);

	  if (ma != mb)
	    {
	      printf ("Mismatch: got %d, expected %d\n", ma, mb);
	      exit (EXIT_FAILURE);
	    }
	}

      class->discard (a);
      bogo_discard (b);

      printf ("Success.\n");
      cmdline.seed++;
    }

  return EXIT_SUCCESS;
}
