/* Copyright (C) 2001 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@msu.edu> on the Internet,
   or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA through
   more mundane means.  
*/

#include <assert.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "test.h"

/* Choose and return an initial random seed based on the current time.
   Based on code by Lawrence Kirby <fred@genesis.demon.co.uk>. */
unsigned 
time_seed (void) 
{
  time_t timeval;
  unsigned char *ptr;
  unsigned seed;
  size_t i;

  timeval = time (NULL);
  ptr = (unsigned char *) &timeval;

  seed = 0;
  for (i = 0; i < sizeof timeval; i++)
    seed = seed * (UCHAR_MAX + 2U) + ptr[i];

  return seed;
}

/* |s| should point to a decimal representation of an integer.
   Returns the value of |s|, if successful, or 0 on failure. */
int 
stoi (const char *s) 
{
  long l = strtol (s, NULL, 10);
  return l >= INT_MIN && l <= INT_MAX ? l : 0;
}

/* Prints |message| on stderr, which is formatted as for printf(), and
   terminates the program unsuccessfully. */
void 
fail (const char *message, ...)
{
  va_list args;

  fprintf (stderr, "%s: ", pgm_name);

  va_start (args, message);
  vfprintf (stderr, message, args);
  va_end (args);

  putchar ('\n');

  exit (EXIT_FAILURE);
}

/* Command-line parser. */

/* Initializes |state| as a command-line options parser.  |args| is a
   null-terminated array of command-line arguments, not including
   program name.  |short_opts| is a string containing the acceptable
   short options; those accepting a required argument are followed by
   a colon (`:').  Long options are passed in |long_opts|, which is
   terminated by an all-zero element. */
void
option_init (struct option_state *state, char **args,
	     const char *short_opts, const struct long_option *long_opts)
{
  assert (state != NULL && args != NULL
	  && short_opts != NULL && long_opts != NULL);

  state->arg_next = args;
  state->short_next = NULL;
  state->short_opts = short_opts;
  state->long_opts = long_opts;
}

/* Parses a short option whose single-character name is pointed to by
   |state->short_next|.  Advances past the option so that the next one
   will be parsed in the next call to option_get().  Sets |*arg| to the
   option's argument, if any.  Returns the short option name. */
static int
handle_short_option (struct option_state *state, char **arg)
{
  char *p;

  assert (state != NULL
	  && state->short_next != NULL && *state->short_next != '\0'
	  && state->short_opts != NULL);

  p = strchr (state->short_opts, *state->short_next);
  if (p == NULL)
    fail ("unknown option `-%c'; use --help for help", *state->short_next);
  state->short_next++;

  if (p[1] == ':')
    {
      if (*state->arg_next == NULL || **state->arg_next == '-')
	fail ("`-%c' requires an argument");

      *arg = *state->arg_next++;
    }

  return *p;
}

/* Parses a long option whose command-line argument is pointed to by
   |*state->arg_next|.  Advances past the option so that the next one
   will be parsed in the next call to option_get().  Sets |*arg| to
   the option's argument, if any.  Returns the option's identifier. */
static int 
handle_long_option (struct option_state *state, char **arg)
{
  const struct long_option *o;	/* Iterator on state->long_opts. */
  char key[16];			/* Option key (name). */
  const char *value;		/* Option value (argument). */

  assert (state != NULL
	  && state->arg_next != NULL && *state->arg_next != NULL
	  && state->long_opts != NULL);

  /* Extra the option's key into |key|
     and its argument, or |NULL| if none, into |value|. */
  {
    const char *p = *state->arg_next + 2;
    const char *q = p + strcspn (p, "=");

    {
      size_t key_len = q - p;
      if (key_len > 15)
	key_len = 15;
      memcpy (key, p, key_len);
      key[key_len] = '\0';
    }
      
    if (*q == '=')
      value = q + 1;
    else
      value = NULL;
  }

  /* Find option in |o|. */
  for (o = state->long_opts; ; o++)
    if (o->key == NULL)
      fail ("unknown option --%s; use --help for help", key);
    else if (!strcmp (key, o->key))
      break;

  /* Make sure option has an argument if it should. */
  if ((value != NULL) != (o->takes_value != 0))
    {
      if (value != NULL)
	fail ("--%s can't take an argument", key);
      else
	fail ("--%s requires an argument", key);
    }

  /* Advance and return. */
  state->arg_next++;
  *arg = (char *) value;
  return o->id;
}

/* Retrieves the next option in the state vector |state|.
   Returns the option's identifier.
   Stores the option's argument, or |NULL| if none, into |*arg|. */
int
option_get (struct option_state *state, char **arg)
{
  assert (state && state->short_opts && state->long_opts);

  /* No argument by default. */
  *arg = NULL;

  /* Deal with left-over short options. */
  if (state->short_next != NULL)
    {
      if (*state->short_next != '\0')
	return handle_short_option (state, arg);
      else
	state->short_next = NULL;
    }

  /* Out of options? */
  if (*state->arg_next == NULL)
    return -1;

  /* Non-option arguments not supported. */
  if ((*state->arg_next)[0] != '-')
    fail ("non-option arguments encountered; use --help for help");
  if ((*state->arg_next)[1] == '\0')
    fail ("unknown option `-'");

  /* Handle the option. */
  if ((*state->arg_next)[1] == '-')
    return handle_long_option (state, arg);
  else
    {
      state->short_next = *state->arg_next + 1;
      state->arg_next++;
      return handle_short_option (state, arg);
    }
}
